@@ -1,5 +1,5 @@
 /* medit.c -- simple multilingual editor.		-*- coding: euc-jp; -*-
-   Copyright (C) 2003, 2004, 2005
+   Copyright (C) 2003, 2004, 2005, 2006, 2007
      National Institute of Advanced Industrial Science and Technology (AIST)
      Registration Number H15PRO112
 
@@ -17,7 +17,7 @@
 
    You should have received a copy of the GNU Lesser General Public
    License along with the m17n library; if not, write to the Free
-   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    02111-1307, USA.  */
 
 /***en
@@ -120,8 +120,6 @@
 #include <X11/Xaw/SmeLine.h>
 #include <X11/Xaw/MenuButton.h>
 
-#define VERSION "1.3.4"
-
 /* Global variables.  */
 
 char *filename;
@@ -865,7 +863,8 @@ show_cursor (XtPointer client_data)
       else
 	{
 	  XtSetArg (arg[0], XtNborderWidth, 1);
-	  XtSetArg (arg[1], XtNlabel, mlanguage_name (sym));
+	  sym = mlanguage_name (sym);
+	  XtSetArg (arg[1], XtNlabel, msymbol_name (sym));
 	  XtSetValues (CurLangWidget, arg, 2);
 	}
       XtSetValues (CurLangWidget, arg, 2);
@@ -2289,16 +2288,16 @@ setup_input_methods (int with_xim, char *initial_input_method)
 
   qsort (input_method_table, num_input_methods, sizeof input_method_table[0],
 	 compare_input_method);
-  mplist_put (minput_driver->callback_list, Minput_status_start,
-	      (void *) input_status);
-  mplist_put (minput_driver->callback_list, Minput_status_draw,
-	      (void *) input_status);
-  mplist_put (minput_driver->callback_list, Minput_status_done,
-	      (void *) input_status);
-  mplist_put (minput_driver->callback_list, Minput_get_surrounding_text,
-	      (void *) surrounding_text_handler);
-  mplist_put (minput_driver->callback_list, Minput_delete_surrounding_text,
-	      (void *) surrounding_text_handler);
+  mplist_put_func (minput_driver->callback_list, Minput_status_start,
+		   M17N_FUNC (input_status));
+  mplist_put_func (minput_driver->callback_list, Minput_status_draw,
+		   M17N_FUNC (input_status));
+  mplist_put_func (minput_driver->callback_list, Minput_status_done,
+		   M17N_FUNC (input_status));
+  mplist_put_func (minput_driver->callback_list, Minput_get_surrounding_text,
+		   M17N_FUNC (surrounding_text_handler));
+  mplist_put_func (minput_driver->callback_list, Minput_delete_surrounding_text,
+		   M17N_FUNC (surrounding_text_handler));
 
   current_input_context = NULL;
   current_input_method = -1;
@@ -2651,8 +2650,8 @@ main (int argc, char **argv)
 	help_exit (argv[0], 0);
       else if (! strcmp (argv[i], "--version"))
 	{
-	  printf ("m17n-edit (m17n library) %s\n", VERSION);
-	  printf ("Copyright (C) 2003 AIST, JAPAN\n");
+	  printf ("m17n-edit (m17n library) %s\n", M17NLIB_VERSION_NAME);
+	  printf ("Copyright (C) 2003, 2004, 2005, 2006, 2007 AIST, JAPAN\n");
 	  exit (0);
 	}
       else if (! strcmp (argv[i], "--geometry"))
