@@ -17,7 +17,7 @@
 
    You should have received a copy of the GNU Lesser General Public
    License along with the m17n library; if not, write to the Free
-   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    02111-1307, USA.  */
 
 /***en
@@ -129,7 +129,8 @@ free_plist (void *object)
   do {
     MPlist *next = plist->next;
 
-    if (MPLIST_KEY (plist) != Mnil && MPLIST_KEY (plist)->managing_key)
+    if (MPLIST_KEY (plist) != Mnil
+	&& MPLIST_KEY (plist)->managing_key)
       M17N_OBJECT_UNREF (MPLIST_VAL (plist));
     M17N_OBJECT_UNREGISTER (plist_table, plist);
     free (plist);
@@ -232,15 +233,9 @@ read_hexadesimal (MStream *st)
 static MPlist *
 read_mtext_element (MPlist *plist, MStream *st, int skip)
 {
-  union {
-    int chars[READ_MTEXT_BUF_SIZE];
-    unsigned char bytes[sizeof (int) * READ_MTEXT_BUF_SIZE];
-  } buffer;
-  unsigned char *bytes = buffer.bytes;
-  int nbytes = sizeof (int) * READ_MTEXT_BUF_SIZE;
-  int *chars = NULL;
-  int nchars = 0;
-  int c, i, j;
+  unsigned char buffer[READ_MTEXT_BUF_SIZE], *buf = buffer;
+  int nbytes = READ_MTEXT_BUF_SIZE;
+  int c, i;
 
   i = 0;
   while ((c = GETC (st)) != EOF && c != '"')
@@ -271,65 +266,34 @@ read_mtext_element (MPlist *plist, MStream *st, int skip)
 
       if (! skip)
 	{
-	  if (is_char && ! chars)
+	  if (i + MAX_UTF8_CHAR_BYTES >= nbytes)
 	    {
-	      chars = buffer.chars;
-	      for (j = i - 1; j >= 0; j--)
-		chars[j] = bytes[j];
-	      nchars = READ_MTEXT_BUF_SIZE;
-	      if (bytes != buffer.bytes)
-		free (bytes);
-	    }
-
-	  if (chars)
-	    {
-	      if (i + 1 >= nchars)
-		{
-		  nchars *= 2;
-		  if (chars == buffer.chars)
-		    {
-		      MTABLE_MALLOC (chars, nchars, MERROR_PLIST);
-		      memcpy (chars, buffer.chars, sizeof (int) * i);
-		    }
-		  else
-		    MTABLE_REALLOC (chars, nchars, MERROR_PLIST);
-		}
-	      chars[i++] = c;
-	    }
-	  else
-	    {
-	      if (i + MAX_UTF8_CHAR_BYTES >= nbytes)
+	      if (buf == buffer)
 		{
 		  nbytes *= 2;
-		  if (bytes == buffer.bytes)
-		    {
-		      MTABLE_MALLOC (bytes, nbytes, MERROR_PLIST);
-		      memcpy (bytes, buffer.bytes, i);
-		    }
-		  else
-		    MTABLE_REALLOC (bytes, nbytes, MERROR_PLIST);
+		  buf = malloc (nbytes);
+		  memcpy (buf, buffer, i);
+		}
+	      else
+		{
+		  nbytes += READ_MTEXT_BUF_SIZE;
+		  buf = realloc (buf, nbytes);
 		}
-	      bytes[i++] = c;
 	    }
+
+	  if (is_char)
+	    i += CHAR_STRING_UTF8 (c, buf);
+	  else
+	    buf[i++] = c;
 	}
     }
 
   if (! skip)
     {
-      MText *mt;
-
-      if (chars)
-	{
-	  mt = mtext__from_data (chars, i, MTEXT_FORMAT_UTF_32, 1);
-	  if (chars != buffer.chars)
-	    free (chars);
-	}	  
-      else
-	{
-	  mt = mtext__from_data (bytes, i, MTEXT_FORMAT_UTF_8, 1);
-	  if (bytes != buffer.bytes)
-	    free (bytes);
-	}
+      MText *mt = mtext__from_data (buf, i, MTEXT_FORMAT_UTF_8,
+				    (buf == buffer));
+      if (buf != buffer)
+	mt->allocated = nbytes;
       MPLIST_SET_ADVANCE (plist, Mtext, mt);
     }
   return plist;
@@ -613,11 +577,17 @@ write_element (MText *mt, MPlist *plist, int indent)
       sprintf (buf, "%d", num);
       PUTS (mt, buf);
     }
-  else if (MPLIST_PLIST_P (plist))
+  else if (MPLIST_PLIST_P (plist)
+	   || MPLIST_NESTED_P (plist))
     {
       MPlist *pl;
       int newline = 0;
 
+      if (MPLIST_NESTED_P (plist))
+	{
+	  write_symbol (mt, MPLIST_KEY (plist));
+	  PUTC (mt, ':');
+	}
       plist = MPLIST_PLIST (plist);
       PUTC (mt, '(');
       if (indent >= 0)
@@ -714,17 +684,12 @@ write_element (MText *mt, MPlist *plist, int indent)
     }
   else 
     {
+      char buf[128];
+
       write_symbol (mt, MPLIST_KEY (plist));
       PUTC (mt, ':');
-      if (MPLIST_NESTED_P (plist))
-	write_element (mt, plist, indent + 1);
-      else
-	{
-	  char buf[128];
-
-	  sprintf (buf, "%04X", (unsigned) MPLIST_VAL (plist));
-	  PUTS (mt, buf);
-	}
+      sprintf (buf, "%04X", (unsigned) MPLIST_VAL (plist));
+      PUTS (mt, buf);
     }
 }
 
@@ -791,6 +756,8 @@ mplist__from_plist (MPlist *plist)
       type = MPLIST_KEY (plist);
       if (type->managing_key && MPLIST_VAL (plist))
 	M17N_OBJECT_REF (MPLIST_VAL (plist));
+      if (type == Mplist)
+	MPLIST_SET_NESTED_P (p);
       MPLIST_SET_ADVANCE (p, key, MPLIST_VAL (plist));
       plist = MPLIST_NEXT (plist);
     }
@@ -819,6 +786,7 @@ mplist__from_alist (MPlist *plist)
       elt = MPLIST_PLIST (plist);
       if (! MPLIST_SYMBOL_P (elt))
 	MERROR (MERROR_PLIST, NULL);
+      MPLIST_SET_NESTED_P (p);
       MPLIST_SET_ADVANCE (p, MPLIST_SYMBOL (elt), MPLIST_NEXT (elt));
       M17N_OBJECT_REF (MPLIST_NEXT (elt));
     }
@@ -924,8 +892,10 @@ mplist__conc (MPlist *plist, MPlist *tail)
   MPLIST_DO (pl, plist);
   MPLIST_KEY (pl) = MPLIST_KEY (tail);
   MPLIST_VAL (pl) = MPLIST_VAL (tail);
-  if (MPLIST_KEY (pl)->managing_key)
+  if (MPLIST_KEY (pl)->managing_key && MPLIST_VAL (pl))
     M17N_OBJECT_REF (MPLIST_VAL (pl));
+  if (MPLIST_NESTED_P (tail))
+    MPLIST_SET_NESTED_P (pl);
   tail = MPLIST_NEXT (tail);
   MPLIST_NEXT (pl) = tail;
   M17N_OBJECT_REF (tail);
@@ -1103,7 +1073,11 @@ mplist_copy (MPlist *plist)
   MPlist *copy = mplist (), *pl = copy;
 
   MPLIST_DO (plist, plist)
-    pl = mplist_add (pl, MPLIST_KEY (plist), MPLIST_VAL (plist));
+    {
+      if (MPLIST_NESTED_P (plist))
+	MPLIST_SET_NESTED_P (pl);
+      pl = mplist_add (pl, MPLIST_KEY (plist), MPLIST_VAL (plist));
+    }
   return copy;
 }
 
@@ -1166,10 +1140,10 @@ mplist_put (MPlist *plist, MSymbol key, void *val)
 /***en
     @brief Get the value of a property in a property list.
 
-    The mplist_get () function searches property list $PLIST
-    from the beginning for a property whose key is $KEY.  If such a
-    property is found, a pointer to its value is returned as the type
-    of <tt>(void *)</tt>.  If not found, @c NULL is returned.
+    The mplist_get () function searches property list $PLIST from the
+    beginning for a property whose key is $KEY.  If such a property is
+    found, its value is returned as the type of <tt>(void *)</tt>.  If
+    not found, @c NULL is returned.
 
     When @c NULL is returned, there are two possibilities: one is the
     case where no property is found (see above); the other is the case
@@ -1180,7 +1154,7 @@ mplist_put (MPlist *plist, MSymbol key, void *val)
     @brief ץѥƥꥹΥץѥƥͤ.
 
     ؿ mplist_get () ϡץѥƥꥹ $PLIST Ϥᤫõơ
-    $KEY Ǥץѥƥ򸫤Ĥ롣ĤСͤؤΥݥ󥿤
+     $KEY Ǥץѥƥ򸫤Ĥ롣ĤСͤ
     <tt>(void *)</tt> ֤Ĥʤ @c NULL ֤
 
     @c NULL ֤äݤˤĤβǽ: 
@@ -1201,6 +1175,67 @@ mplist_get (MPlist *plist, MSymbol key)
 
 /*=*/
 
+/***en
+    @brief Set the value (function pointer) of a property in a property list.
+
+    The mplist_put_func () function is similar to mplist_put () but for
+    setting function pointer $FUNC in property list $PLIST for key
+    $KEY.  $KEY must not be a managing key.  */
+
+/***ja
+    @brief ץѥƥꥹΥץѥƥ˴ؿݥ󥿤Ǥͤꤹ.
+
+    ؿ mplist_put_func () ϴؿ mplist_put () Ʊ͡ץѥƥꥹ $PLIST
+    ǥ $KEY Ǥץѥƥͤꤹ롣âͤϴؿݥ
+    $FUNC Ǥ롣$KEY ϴǤäƤϤʤʤ  */
+
+
+/***
+    @seealso
+    mplist_put (), M17N_FUNC ()  */
+
+MPlist *
+mplist_put_func (MPlist *plist, MSymbol key, M17NFunc func)
+{
+  if (key == Mnil || key->managing_key)
+    MERROR (MERROR_PLIST, NULL);
+  MPLIST_FIND (plist, key);
+  MPLIST_KEY (plist) = key;
+  MPLIST_FUNC (plist) = func;
+  MPLIST_SET_VAL_FUNC_P (plist);
+  if (! plist->next)
+    MPLIST_NEW ((plist)->next);
+  return plist;
+}
+
+/*=*/
+
+/***en
+    @brief Get the value (function pointer) of a property in a property list.
+
+    The mplist_get_func () function is similar to mplist_get () but for
+    getting a function pointer from property list $PLIST by key $KEY.  */
+
+/***ja
+    @brief ץѥƥꥹȤץѥƥδؿݥ󥿤Ǥͤ.
+
+    ؿ mplist_get_func () ϴؿ mplist_get () Ʊͤˡץѥƥ
+     $PLIST ǥ $KEY Ǥץѥƥ͡âؿݥ󥿡
+    롣 */
+
+
+/***
+    @seealso
+    mplist_get () */
+M17NFunc
+mplist_get_func (MPlist *plist, MSymbol key)
+{
+  MPLIST_FIND (plist, key);
+  return (MPLIST_TAIL_P (plist) ? NULL : MPLIST_FUNC (plist));
+}
+
+/*=*/
+
 /***en
     @brief Add a property at the end of a property list.
 
@@ -1279,6 +1314,8 @@ mplist_push (MPlist *plist, MSymbol key, void *val)
   MPLIST_NEW (pl);
   MPLIST_KEY (pl) = MPLIST_KEY (plist);
   MPLIST_VAL (pl) = MPLIST_VAL (plist);
+  if (MPLIST_NESTED_P (plist))
+    MPLIST_SET_NESTED_P (pl);
   MPLIST_NEXT (pl) = MPLIST_NEXT (plist);
   plist->next = pl;
   if (val && key->managing_key)
@@ -1295,7 +1332,7 @@ mplist_push (MPlist *plist, MSymbol key, void *val)
 
     The mplist_pop () function removes a property at the beginning of
     property list $PLIST.  As a result, the second key and value of
-    the original $PLIST become the first of those of the new $PLIST.
+    the $PLIST become the first ones.
 
     @return
     If the operation was successful, this function return the value of
@@ -1303,9 +1340,8 @@ mplist_push (MPlist *plist, MSymbol key, void *val)
 /***ja
     @brief ץѥƥꥹȤƬץѥƥ.
 
-    ؿ mplist_pop () ϥץѥƥꥹ $PLIST 
-    ƬΥץѥƥ롣̤Ȥơ $PLIST 2ܤΥͤ 
-    $PLIST ƬΥͤˤʤ롣
+    ؿ mplist_pop () ϥץѥƥꥹ $PLIST ƬΥץѥƥ
+    롣̤Ȥơ2ܤΥͤƬΥͤˤʤ롣
 
     @return 
     СδؿϺ줿ץѥƥ֤ͤǤʤ
